from __future__ import annotations

import abc
from dataclasses import dataclass
from importlib import metadata

from loguru import logger
from omu import Omu
from omu.helper import Coro
from omu_chat import Channel, Chat, Provider, Room

type ChatServiceFactory = Coro[[], ChatService]


@dataclass(frozen=True, slots=True)
class FetchedRoom:
    room: Room
    create: ChatServiceFactory


class ProviderService(abc.ABC):
    @abc.abstractmethod
    def __init__(self, omu: Omu, chat: Chat): ...

    @property
    @abc.abstractmethod
    def provider(self) -> Provider: ...

    @abc.abstractmethod
    async def fetch_rooms(self, channel: Channel) -> list[FetchedRoom]: ...

    @abc.abstractmethod
    async def is_online(self, room: Room) -> bool: ...


class ChatService(abc.ABC):
    @property
    @abc.abstractmethod
    def room(self) -> Room: ...

    @property
    @abc.abstractmethod
    def closed(self) -> bool: ...

    @abc.abstractmethod
    async def run(self): ...

    async def start(self):
        try:
            await self.run()
        except Exception as e:
            logger.opt(exception=e).error(f"Error starting chat for {self.room.key()}")
        finally:
            await self.stop()

    @abc.abstractmethod
    async def stop(self): ...


def retrieve_services():
    entry_points = metadata.entry_points(group="omu_chatprovider.services")

    services: list[type[ProviderService]] = []
    for entry_point in entry_points:
        service = entry_point.load()
        assert issubclass(
            service, ProviderService
        ), f"{service} is not a ProviderService"
        services.append(service)

    return services


__all__ = [
    "ChatService",
    "FetchedRoom",
    "ProviderService",
    "retrieve_services",
]



import asyncio
import time
import traceback
import typing


class Tasks:
    def __init__(self, loop: asyncio.AbstractEventLoop) -> None:
        self.loop = loop
        self.tasks: list[asyncio.Task] = []
        self.times: dict[str, float] = {}
        self.times_max: dict[str, float] = {}

    def terminate(self):
        for task in self.tasks:
            task.cancel()

    def create_task(self, coro: typing.Coroutine):
        name = coro.__name__

        async def wrapper():
            start_time = time.time()
            try:
                await coro
            except BaseException as e:
                traceback.print_exc()
                raise Exception(f"Task {name} failed") from e
            finally:
                self.times[name] = time.time() - start_time
                self.times_max[name] = max(
                    self.times_max.get(name, 0), self.times[name]
                )

        self.tasks.append(asyncio.create_task(wrapper()))



from omu import Plugin

from .version import VERSION

__version__ = VERSION
__all__ = ["plugin"]


def get_client():
    from .chatprovider import omu

    return omu


plugin = Plugin(
    get_client,
    isolated=True,
)



from __future__ import annotations

import json
import re
import time
from collections.abc import Callable, Coroutine

import aiohttp
from loguru import logger
from omu.omu import Omu
from omu_chat import Provider

HTTP_REGEX = r"(https?://)?(www\.)?"
URL_NORMALIZE_REGEX = (
    r"(?P<protocol>https?)?:?\/?\/?"
    r"(?P<domain>[^.]+\.[^\/]+)"
    r"(?P<path>[^?#]+)?"
    r"(?P<query>.+)?"
)


def get_session(omu: Omu, provider: Provider) -> aiohttp.ClientSession:
    user_agent = json.dumps(
        [
            "omu_chatprovider",
            {
                "id": provider.id.key(),
                "version": provider.version,
                "repository_url": provider.repository_url,
            },
        ]
    )
    session = aiohttp.ClientSession(
        loop=omu.loop,
        headers={"User-Agent": user_agent},
    )
    return session


def normalize_url(url: str) -> str:
    match = re.match(URL_NORMALIZE_REGEX, url)
    if match is None:
        raise ValueError(f"Invalid URL: {url}")
    protocol = match.group("protocol") or "https"
    domain = match.group("domain")
    path = match.group("path") or ""
    query = match.group("query") or ""
    return f"{protocol}://{domain}{path}{query}"


def assert_none[T](value: T | None, message: str) -> T:
    if value is None:
        raise ValueError(message)
    return value


def timeit[**P, R](func: Callable[P, R]) -> Callable[P, R]:
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        start = time.time()
        result = func(*args, **kwargs)
        logger.info(f"{func.__name__} took {time.time() - start} seconds")
        return result

    return wrapper


def timeit_async[**P, R](
    func: Callable[P, Coroutine[None, None, R]],
) -> Callable[P, Coroutine[None, None, R]]:
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        start = time.time()
        result = await func(*args, **kwargs)
        logger.info(f"{func.__name__} took {time.time() - start} seconds")
        return result

    return wrapper


class Traverser[T]:
    def __init__(self, value: T | None = None):
        self.value = value

    def map[V](self, func: Callable[[T], V], default: V | None = None) -> Traverser[V]:
        if self.value is None:
            return TRAVERSE_NONE
        value = func(self.value)
        if value is None:
            return Traverser(default)
        return Traverser(value)

    def get(self, default: T | None = None) -> T | None:
        if self.value is None:
            return default
        return self.value

    def unwrap(self) -> T:
        if self.value is None:
            raise ValueError("Value is None")
        return self.value

    def is_none(self) -> bool:
        return self.value is None


TRAVERSE_NONE = Traverser(None)


def traverse[T](value: T | None) -> Traverser[T]:
    """
    Traverse a value that may be None.
    """
    return Traverser(value)
